/***************************************************************************
*   Copyright (C) 2003 by Sandro Giessl                                   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#ifndef ALLOYCLIENT_H
#define ALLOYCLIENT_H

#include <kwin/kwinbutton.h>
#include <kwin/client.h>

class QGridLayout;
class QSpacerItem;
class QPoint;

namespace KWinInternal {

enum ButtonType {
    ButtonHelp=0,
    ButtonMax,
    ButtonMin,
    ButtonClose,
    ButtonMenu,
    ButtonSticky,
    ButtonTypeCount
};

enum BorderMode {
    MaximizedHorizontal=0,
    MaximizedVertical,
    MaximizedFull,
    Normal
};

class AlloyHandler: public QObject
{
public:
    AlloyHandler(); ///< Constructor.
    ~AlloyHandler(); ///< Destructor.
    void reset(); ///< Reset the handler.

    static bool initialized() { return m_initialized; };

    static bool titleShadow() { return m_titleShadow; };
    static bool shrinkBorders() { return m_shrinkBorders; };
    static int  borderSize() { return m_borderSize; };
private:
    void readConfig(); ///< Read in the configuration file.

    static bool m_titleShadow;
    static bool m_shrinkBorders;
    static int  m_borderSize;

    static bool m_initialized;
};

class AlloyButton : public KWinInternal::KWinButton
{
public:
    AlloyButton(Client *parent=0, const char *name=0, const QString &tip=NULL, ButtonType type = ButtonHelp); ///< Constructor.
    ~AlloyButton(); ///< Destructor.

    void setSticky(bool sticky) { isSticky = sticky; repaint(false); }
    void setMaximized(bool maximized) { isMaximized = maximized; repaint(false); }
    QSize sizeHint() const; ///< Return size hint.
    int lastMousePress() const { return m_lastMouse; };
    void reset() { repaint(false); };
    Client* client() { return m_client; }

private:
    void enterEvent(QEvent *e); ///< Mouse has entered the button.
    void leaveEvent(QEvent *e); ///< Mouse has left the button.
    void mousePressEvent(QMouseEvent *e); ///< Button has been pressed.
    void mouseReleaseEvent(QMouseEvent *e); ///< Button has been released.
    void drawButton(QPainter *painter); ///< Draw the button.

private:
    Client *m_client;
    int m_lastMouse;

    ButtonType m_type;
    bool hover;
    bool isSticky, isMaximized;
};


class AlloyClient : public KWinInternal::Client
{
    Q_OBJECT
public:
    AlloyClient(Workspace *ws, WId w, QWidget *parent = 0, const char *name = 0); ///< Constructor.
    ~AlloyClient(); ///< Destructor.
private:
    void addButtons(QBoxLayout* layout, const QString& buttons); ///< Add buttons to title layout.

    void resizeEvent(QResizeEvent *e); ///< Window is being resized.
    void paintEvent(QPaintEvent *e); ///< Repaint the window.

    void mouseDoubleClickEvent(QMouseEvent *e); ///< Doubleclick on title.
    void captionChange(const QString &name); ///< The title has changed.
    void maximizeChange(bool m); ///< Maximized state has changed.
    void stickyChange(bool s); ///< Sticky state has changed.
    void iconChange(); ///< The title has changed.
    void activeChange(bool a); ///< window active state has changed.
    MousePosition mousePosition(const QPoint &point) const; ///< Return logical mouse position.

private slots:
    void maxButtonPressed(); ///< Max button was pressed.
    void menuButtonPressed(); ///< Menu button was pressed (popup the menu).

private:
    AlloyButton *m_button[ButtonTypeCount];
    QSpacerItem *m_titleBar, *m_leftSpacer, *m_rightSpacer, *m_topSpacer, *m_bottomSpacer, *m_leftButtonSpacer, *m_leftButtonSpacer2, *m_rightButtonSpacer, *m_rightButtonSpacer2;
    BorderMode m_borderMode;
};

} // namespace KWinInternal

#endif // ALLOYCLIENT_H
