/***************************************************************************
*   Copyright (C) 2003 by Sandro Giessl                                   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include <kconfig.h>
#include <klocale.h>
#include <kglobal.h>
#include <qcheckbox.h>
#include <qslider.h>
#include <qwhatsthis.h>

#include "alloyconfig.h"
#include "configdialog.h"

AlloyConfig::AlloyConfig(KConfig* config, QWidget* parent)
    : QObject(parent), m_config(0), m_dialog(0)
{
    // create the configuration object
    m_config = new KConfig("kwinalloyrc");
    KGlobal::locale()->insertCatalogue("kwin_alloy_config");

    // create and show the configuration dialog
    m_dialog = new ConfigDialog(parent);
    m_dialog->show();

    // load the configuration
    load(config);

    // setup the connections
    connect(m_dialog->m_titleShadow, SIGNAL(toggled(bool)),
            this, SIGNAL(changed()));
    connect(m_dialog->m_shrinkBorders, SIGNAL(toggled(bool)),
            this, SIGNAL(changed()));
    connect(m_dialog->m_borderSize, SIGNAL(valueChanged(int)),
            this, SIGNAL(changed()));

}

AlloyConfig::~AlloyConfig()
{
    if (m_dialog) delete m_dialog;
    if (m_config) delete m_config;
}

void AlloyConfig::load(KConfig*)
{
    m_config->setGroup("General");

    bool titleShadow = m_config->readBoolEntry("TitleShadow", true);
    m_dialog->m_titleShadow->setChecked(titleShadow);
    bool shrinkBorders = m_config->readBoolEntry("ShrinkBorders", true);
    m_dialog->m_shrinkBorders->setChecked(shrinkBorders);
    int borderSize = m_config->readNumEntry("BorderSize", 5);
    m_dialog->m_borderSize->setValue(borderSize);
}

void AlloyConfig::save(KConfig*)
{
    m_config->setGroup("General");

    m_config->writeEntry("TitleShadow", m_dialog->m_titleShadow->isChecked() );
    m_config->writeEntry("ShrinkBorders", m_dialog->m_shrinkBorders->isChecked() );
    m_config->writeEntry("BorderSize", m_dialog->m_borderSize->value() );
    m_config->sync();
}

void AlloyConfig::defaults()
{
    m_dialog->m_titleShadow->setChecked(true);
    m_dialog->m_shrinkBorders->setChecked(true);
    m_dialog->m_borderSize->setValue(5);
}

//////////////////////////////////////////////////////////////////////////////
// Plugin Stuff                                                             //
//////////////////////////////////////////////////////////////////////////////

extern "C"
{
    QObject* allocate_config(KConfig* config, QWidget* parent) {
        return (new AlloyConfig(config, parent));
    }
}

#include "alloyconfig.moc"
